#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// bytecode record types

#define COMMAND     0x11
#define ROUTINE     0x12
#define VARIABLE    0x13
#define NORMAL      0x14
#define BUFFER      0x15
#define KEYTABLE    0x17
#define ONKEY       0x18
#define STRINGS     0x19
#define WINDOW      0x1A
#define COLOR       0x1B 

#define VOLATILE    0x20

// file header 
typedef struct eel_chek {
  char  name[6];                // literal EEL\x1aF where F is version
  int   strings;                // size of constant string area
} eel_chek;

// subroutine header
typedef struct eel_head {
  int   size;                   // size of subroutine
  int   refs;                   // size of external references
  int   xref;                   // number of external references
  int   parm;                   // number of parameters
  int   locals;                 // number of local variables
} eel_head;

eel_chek chek;
eel_head head;

char data[20000];               // read buffer

typedef struct keytab {
  int    count;                 // number of commands referencing this table
  int    table;                 // table sequence number
} keytab;      

keytab keytabs[256] = { 0 };

// binary tree lookup table (probably overkill)
typedef struct node {
  struct node *left;
  struct node *rite;
  char name[1];                 // full name is allocated after this
} node;      

node *last, *root = NULL;                  

// lookup a name in the binary tree of external references
node *lookup(char *name)
  {     
    int  cmp;
    node *next = root;
    
    while (next != NULL)
      {         
        last = next;
        cmp = strcmp(next->name, name);
        if (cmp == 0)
            break;
        if (cmp > 0)
            next = next->rite;
        else
            next = next->left;
      }
    return next;
  }

// enter an external reference into the table
void enter_name(char *name)
  {            
    node *leaf;
    
    if (root == NULL)
      {
        root = (node*)malloc(sizeof(node) + strlen(name));
        root->left = NULL;
        root->rite = NULL;
        strcpy(root->name, name);
      }
    else
      {                      
        if (lookup(name) == 0)
          {
            leaf = (node*)malloc(sizeof(node) + strlen(name));
            leaf->left = NULL;
            leaf->rite = NULL;
            strcpy(leaf->name, name);
            if (strcmp(last->name, name) > 0)
                last->rite = leaf;
            else
                last->left = leaf;
          }
      }
  }

// read a zero terminated string
// return a pointer past the end of the string
char *read_text(FILE *f, char *c)
  {
    int n;

    do
      {
        n = fread(c++, 1, 1, f);
      } while (n == 1 && c[-1]);
    return c;
  }

// read a bytecode file looking for external references
// enter each reference into a lookup table
void read_file(FILE *fi)
  {
    char byte, *cp;
    int  i, j, l;
    unsigned char c;

    if (chek.strings > 20000)
      {
        fprintf(stderr, "string area too large\n");
        exit(3);
      }
    do {
      if (fread(&byte, 1, 1, fi) != 1)
          return;       
      switch (byte) {
          case COMMAND:
          case ROUTINE:
              read_text(fi, data);
              fread((char*)&head, sizeof(head), 1, fi);
			  fread(data, 1, head.size - 4, fi);
			  fread(data, 2, head.xref, fi);
			  for (i = 0; i < head.xref; i++)
				{
				  read_text(fi, data);
				  enter_name(data);
				}
			  break;
          case VARIABLE:
              fread(data, 1, 1, fi);
              cp = read_text(fi, data + 1);
              fread(cp, 1, 2, fi);
              break;
          case NORMAL:
		  case BUFFER:
          case WINDOW:
			  fread(data, 1, 1, fi);
			  cp = read_text(fi, data + 1);
			  cp += fread(cp, 1, 2, fi);
			  i = ((int*)cp)[-1];
			  while (i > 0)
				{
				  j = i;
				  if (j > data + sizeof(data) - cp)
					  j = data + sizeof(data) - cp;
				  fread(cp, 1, j, fi);
				  cp = data;
				  i -= j;
				}
			  break;
		  case KEYTABLE:
			  read_text(fi, data);
              break;
          case ONKEY:
              fread(data, 1, 3, fi);
              keytabs[data[0] & 0xFF].count++;
              break;
          case STRINGS:
              break;
          case COLOR:
			  fread(data, 1, 1, fi);
			  cp = read_text(fi, data + 1);
			  fread(cp, 1, 1, fi);
              c = *cp++;
			  cp += fread(cp, 1, c, fi);
              break;
          default:
              
              fprintf(stderr, "can't do %04X %02X\n",(int)ftell(fi), byte);
              exit(2);
      }
    } while (byte != STRINGS);
  }

// copy a bytecode file skipping all variables that are not referenced
// by any command or subroutine.
// if a variable is volatile, keep it even if it is not referenced
void copy_file(FILE *fi, FILE *fo)
  {
    char *cp;
    int  i, j, l;
    int  curtab;
    unsigned char c;

    for (i = 0, j = 0; i < 256; i++)
        if (keytabs[i].count > 0)
            keytabs[i].table = j++;
    curtab = 0;    
	for (;;)
	 {
	  if (fread(data, 1, 1, fi) != 1)
		  return;
	  switch (*data) {
		  case COMMAND:
		  case ROUTINE:
              cp = read_text(fi, data + 1);
              fwrite(data, 1, cp - data, fo);
              fread((char*)&head, 1, sizeof(head), fi);
              fwrite((char*)&head, 1, sizeof(head), fo);
			  fread(data, 1, head.size - 4, fi);
              fwrite(data, 1, head.size - 4, fo);
              fread(data, 1, head.refs, fi);
              fwrite(data, 1, head.refs, fo);
              break;
		  case VARIABLE:
              fread(data + 1, 1, 1, fi);
              cp = read_text(fi, data + 2);
              cp += fread(cp, 1, 2, fi);         
              if ((data[1] & VOLATILE) || lookup(data + 2))
				  fwrite(data, 1, cp - data, fo);
              break;
          case NORMAL:
          case BUFFER:
          case WINDOW:
			  fread(data + 1, 1, 1, fi);
			  cp = read_text(fi, data + 2);
			  cp += fread(cp, 1, 2, fi);
			  i = ((int*)cp)[-1];
			  c = (data[1] & VOLATILE) || lookup(data + 2) != NULL;
			  while (i > 0)
				{
				  j = i;
				  if (j > data + sizeof(data) - cp)
					  j = data + sizeof(data) - cp;
				  fread(cp, 1, j, fi);
				  if (c)
					  fwrite(data, 1, cp - data + j, fo);
				  cp = data;
				  i -= j;
				}
			  break;
		  case KEYTABLE:
			  cp = read_text(fi, data + 1);
              if (keytabs[curtab++].count > 0)
                  fwrite(data, 1, cp - data, fo);
              break;
          case ONKEY:
              fread(data + 1, 1, 3, fi);
              data[1] = keytabs[data[1] & 0xFF].table;
              fwrite(data, 1, 4, fo);
			  break;
          case STRINGS:
              fread(data + 1, 1, chek.strings, fi);
              fwrite(data, 1, chek.strings + 1, fo);
			  break;
		  case COLOR:
			  fread(data, 1, 1, fi);
			  cp = read_text(fi, data + 2);
			  fread(cp, 1, 1, fi);
              c = *cp++;
			  cp += fread(cp, 1, c, fi);
              if ((data[1] & VOLATILE) || lookup(data + 2))
				  fwrite(data, 1, cp - data, fo);
              break;
	  }
	}
  }

// strip unused variables generated by eel.h etc. from a bytecode file
// this only works on type "F" files (version 6.5x)
int main(int argc, char *argv[])
  {
    FILE *fi, *fo;

    printf("\nStrip ver 1.0\n\n");
    printf("   bytecode stripper\n");
    printf("   by Rob Duff (604) 988-1824\n");
    if (argc == 3 &&
        (fi = fopen(argv[1], "rb")) != NULL &&
        (fo = fopen(argv[2], "wb")) != NULL)
      {
        fread((char*)&chek, 1, sizeof(chek), fi);
        if (strcmp(chek.name, "EEL\x1A""F") == 0)
          {
            read_file(fi);
            fwrite((char*)&chek, 1, sizeof(chek), fo);
			fseek(fi, (long)sizeof(chek), SEEK_SET);
            copy_file(fi, fo);
            fclose(fo);
            return 0;
          }
        fprintf(stderr, "Not a EEL 6.5x bytecode file\n");
      }
    else if (argc != 3)
        fprintf(stderr, "Usage: strip <infile> <outfile>\n");
    else if (fi == NULL)
        fprintf(stderr, "Can't open %s\n", argv[1]);
    else
        fprintf(stderr, "Can't create %s\n", argv[2]);
    return 1;
  }